#ifndef __VRAYMESHGEOM_H__
#define __VRAYMESHGEOM_H__

#include "max.h"
#include <bmmlib.h>
#include "iparamm2.h"
#include "render.h"  
#include "texutil.h"
#include "gizmo.h"
#include "gizmoimp.h"
#include "istdplug.h"
#include "notify.h"
#include "vassert.h"
#include "vraygeom.h"
#include "rayserver.h"

#include "mesh_file.h"

#include "plugman.h"
#include "vrayplugins.h"
#include "factory.h"

//************************************************************
// #defines
//************************************************************

#define	PLUGIN_CLASSID Class_ID(0x2d3f496e, 0x5517642f)

#define STR_CLASSNAME _T("VRayMeshGeom")
#define STR_INTERNALNAME _T("VRayMeshGeom")
#define STR_LIBDESC _T("VRay SDK VRayMeshGeom example")
#define STR_DLGTITLE _T("VRayMeshGeom Parameters")

#define REFNO_PBLOCK 0

// Paramblock2 parameter list
enum {
	pb_numSpheres,
	pb_radius,
	pb_numSegments,
	pb_boxSize,
	pb_animated,
};

//************************************************************
// The mesh generator

struct VoxelInfo {
	VR::InterlockedCounter loadedCount;
	int isLoaded;

	VoxelInfo(void) {
		loadedCount.set(0);
		isLoaded=false;
	}
};

class MeshTest: public VR::MeshInterface {
public:
	MeshTest(void);
	~MeshTest(void);

	void init(int numSpheres, float sphereSize, int sphereSegments, float bboxSize, float time, int isMoving, VR::RayServerInterface *rayserver, VR::ThreadManager *threadman);

	// From MeshInterface
	int getNumVoxels(void);
	VR::Box getVoxelBBox(int i);
	uint32 getVoxelFlags(int i);
	VR::MeshVoxel* getVoxel(int i, uint64 *memUsage) override;
	VR::MeshVoxel* getVoxel(int i, int timeFlags, uint64 *memUsage, VR::GetContextDataCallbackBase *ctxtDataCallback) override {
		// For now we don't support time flags so just forward to the other getVoxel method
		return getVoxel(i,memUsage);
	}
	void releaseVoxel(VR::MeshVoxel *voxel, uint64 *memUsage);

	int getNumSpheres(void) { return numSpheres; }

protected:
	float size;
	int subdivs;
	float bboxSize;
	float time;
	int isMoving;

	int numSpheres;
	VR::MeshVoxel *voxels;
	VoxelInfo *voxelInfo;
	VR::FastCriticalSection csect;
	VR::RayServerInterface *rayserver;
	VR::ThreadManager *threadman;

	void buildVoxel(int voxelIndex, VR::MeshVoxel &mesh);
	void setFace(VR::FaceTopoData &f, int v0, int v1, int v2);
	void getRandomSphere(int seed, float &radius, VR::Vector &center, float time);
};

//************************************************************
// The VRayMeshGeom 3dsmax object

class VRayMeshGeom: public GeomObject, public VR::VRenderObject {
	// An empty dummy mesh returned from GetRenderMesh()
	Mesh mesh;
	MeshTest meshGen;
	int isMoving;

	static Factory factory; // factory to hold the parameters of shading plugins
	static PluginManager *plugman; // plugin manager to hold the shading plugins
	static int numRenderPlugins; // The number of VRayMeshGeom plugins that use the plugin manager; used to call
		// cleanShaders() when the last plugin receives RenderEnd().
	static int numFramePlugins;

	VR::VRayPlugin* newPlugin(const tchar *type, VR::ProgressCallback *prog); // create a new plugin of the given type
	void clearShaders(); // clear the shading plugins and their parameters

	// Create a material for the instance with the given ID.
	VR::VRayPlugin* createSomeMaterial(int renderID, VR::ProgressCallback *prog);

	// Create a default plugin manager, if one does not exist already
	void createPluginManager(VR::ProgressCallback *prog);

	// Called when 3ds Max shuts down - we need to unload the V-Ray shader DLLs
	static void maxShutdown(void *param, NotifyInfo *info);
public:
	IParamBlock2 *pblock2;
	IParamMap2 *pmap;
	static IObjParam *ip;

	// Snap suspension flag (TRUE during creation only)
	BOOL suspendSnap;
				
	float simple;
 	int extDispFlags;

	VRayMeshGeom(void);
	~VRayMeshGeom(void);
	
	// From BaseObject
	int IsRenderable() { return true; }
	int HitTest(TimeValue t, INode* inode, int type, int crossing, int flags, IPoint2 *p, ViewExp *vpt);
	void Snap(TimeValue t, INode* inode, SnapInfo *snap, IPoint2 *p, ViewExp *vpt);
	void SetExtendedDisplay(int flags);
	int Display(TimeValue t, INode* inode, ViewExp *vpt, int flags);
	CreateMouseCallBack* GetCreateMouseCallBack();

	void BeginEditParams( IObjParam *ip, ULONG flags,Animatable *prev);
	void EndEditParams( IObjParam *ip, ULONG flags,Animatable *next);
	void InvalidateUI();

	// From Object
	ObjectState Eval(TimeValue time);

	void InitNodeName(TSTR& s) { s=STR_CLASSNAME; }
	ObjectHandle ApplyTransform(Matrix3& matrix) { return this; }
	Interval ObjectValidity(TimeValue t);

	// We don't convert to anything
	int CanConvertToType(Class_ID obtype) { return FALSE; }
	Object* ConvertToType(TimeValue t, Class_ID obtype) { vassert(0);return NULL; }
	
	void GetWorldBoundBox(TimeValue t, INode *mat, ViewExp *vpt, Box3& box);
	void GetLocalBoundBox(TimeValue t, INode *mat, ViewExp *vpt, Box3& box);
	void GetDeformBBox(TimeValue t, Box3 &b, Matrix3 *tm, BOOL useSel);
	int DoOwnSelectHilite()	{ return 1; }

	// From GeomObject
	Mesh* GetRenderMesh(TimeValue t, INode *inode, View& view, BOOL& needDelete);

	// Animatable methods
	void DeleteThis() { delete this; }
	Class_ID ClassID() { return PLUGIN_CLASSID; }
	void GetClassName(TSTR& s) { s=_T("VRayMeshGeom"); }
	int IsKeyable() { return 0; }
	void* GetInterface(ULONG id);
	void ReleaseInterface(ULONG id, void *ip);
	
	// Direct paramblock access
	int	NumParamBlocks() { return 1; }	
	IParamBlock2* GetParamBlock(int i) { return pblock2; }
	IParamBlock2* GetParamBlockByID(BlockID id) { return (pblock2->ID() == id) ? pblock2 : NULL; }

	int NumSubs() { return 1; }  
	Animatable* SubAnim(int i);
	TSTR SubAnimName(int i);

	// From ref
 	int NumRefs() { return 1; }
	RefTargetHandle GetReference(int i);
	void SetReference(int i, RefTargetHandle rtarg);

#if GET_MAX_RELEASE(VERSION_3DSMAX) < 8900
	RefTargetHandle Clone(RemapDir& remap=NoRemap());
#else
	RefTargetHandle Clone(RemapDir& remap=DefaultRemapDir());
#endif
	RefResult NotifyRefChanged(NOTIFY_REF_CHANGED_ARGS);

	// From VRenderObject
	int init(const ObjectState &os, INode *node, VR::VRayCore *vray);
	VR::VRenderInstance* newRenderInstance(INode *node, VR::VRayCore *vray, int renderID);
	void deleteRenderInstance(VR::VRenderInstance *ri);
	void renderBegin(TimeValue t, VR::VRayCore *vray);
	void renderEnd(VR::VRayCore *vray);
	void frameBegin(TimeValue t, VR::VRayCore *vray);
	void frameEnd(VR::VRayCore *vray);

	// From MeshInterface
	int getNumVoxels(void) { return 1; }
	uint32 getVoxelFlags(int i) { return MVF_GEOMETRY_VOXEL; }
	VR::Box getVoxelBBox(int i); // Return the bounding box of the mesh
	VR::MeshVoxel* getVoxel(int i, uint64 *memUsage); // Generate the mesh itself
	void releaseVoxel(VR::MeshVoxel *voxel, uint64 *memUsage); // Free any memory allocated for the voxel

	// Other methods
	void draw(TimeValue t, INode *node, ViewExp *vpt);

	VR::MeshInterface* getMeshInterface(void) { return &meshGen; }
};

class VRayMeshGeomCreateCallBack: public CreateMouseCallBack {
	VRayMeshGeom *sphere;
	IPoint2 sp0;
	Point3 p0;
public:
	int proc(ViewExp *vpt, int msg, int point, int flags, IPoint2 m, Matrix3& mat);
	void SetObj(VRayMeshGeom *obj) { sphere=obj; }
};

#endif